
#include "x.h"
#include "xpex.h"
#include <math.h>
#include "book_utils.h"
#include "externs.h"
#include "defs.h"

XtAppContext  app;
Display      *dpy;
Window        window;
Drawable      drawable;
Atom          delete_window_atom;

PEXRenderer           renderer;
PEXRendererAttributes rattrs;

PEXStructure top_struct, sphere;

Widget toplevel, canvas;

PEXColorSpecifier colorSpec;

Boolean colorTypeIndexed, colorTypeRGB,
        colorTypeHSV, colorTypeHLS, colorTypeCIE;

PEXDepthCueEntry        depth_cue_rep;
PEXReflectionAttributes    refl_props;

int cur_light = SPOT;
int cur_shading = PEXSurfaceInterpColor;
int cur_refl = PEXReflectionSpecular;
int cur_hlhsr_mode = PEXHLHSRZBuffer;

int light_switches[4] = { 1, 1, 1, 0 };
PEXLightEntry lights[4];
int num_on=3, num_off=1;
PEXTableIndex light_on[4] = { 1, 2, 3, 0 };
PEXTableIndex light_off[4] = { 4, 0, 0, 0 };

/* The viewing parameters. */
PEXCoord         view_ref_pt = { 0.5,0.5,0.5};
PEXVector        view_up_vec = { 0,1,0};
PEXVector        view_plane_normal = { 1,1,1};
PEXCoord2D       view_window[2];
PEXNPCSubVolume  viewport;
double           view_plane, front_plane, back_plane;
PEXCoord         prp;
int              perspective;
PEXViewEntry     view;


int islighton( light )
int light;
{
    register int i;
    for( i=0; i<4; i++ )
    if( light_on[i] == light )
        return 1;
    return 0;
}


#define NORMALIZE( _v )                          \
{                                                \
    double mag;                                  \
                                                 \
    mag = 1.0 / sqrt((_v).x * (_v).x             \
          + (_v).y * (_v).y + (_v).z * (_v).z);  \
    (_v).x *= mag, (_v).y *= mag, (_v).z *= mag; \
}

main( argc, argv )
int        argc;
char    *argv[];
{
    XVisualInfo            vis_info;
    XStandardColormap        cmap_info;
    PEXColorApproxEntry        capx_info;
    XSetWindowAttributes    wattrs;
    PEXExtensionInfo        *pexinfo;
    char            err[80];
    extern int PrintDefaultError();
    XColor background;

    unsigned long   pcMask[3];
    PEXPCAttributes pcattrs;

    PEXColor      color;
    PEXCoord      pln[2];

    /* Initialize light table and depth cueing table. */
    initialize_light_table();

    if( argc > 1 ) {
    float stof();
    register int i;
    int pos = 1;
    char ltype[3];

    for( i=0; i<4; i++ ) {
        light_switches[i] = light_on[i] = light_off[i] = 0;
    }

    num_on = stof( argv[pos+2] );
    pos += 3;
    for( i=0; i<num_on; i++ ) {
        sscanf( argv[pos], "%s", &ltype );
        if( !strcmp( ltype, "Amb" ) ) {
            light_on[i] = 1;
            light_switches[AMBIENT] = 1;
            SET_COLOR( stof(argv[pos+2]), stof(argv[pos+3]), stof(argv[pos+4]),
                        lights[AMBIENT].color.value );
            pos+=5;
        }
        else if( !strcmp( ltype, "Dir" ) ) {
            light_on[i] = 2;
            light_switches[DIRECTIONAL] = 1;
            lights[DIRECTIONAL].direction.x = stof( argv[pos+1] );
            lights[DIRECTIONAL].direction.y = stof( argv[pos+2] );
            lights[DIRECTIONAL].direction.z = stof( argv[pos+3] );
            NORMALIZE(lights[DIRECTIONAL].direction);
            SET_COLOR( stof(argv[pos+5]), stof(argv[pos+6]), stof(argv[pos+7]),
                        lights[DIRECTIONAL].color.value );
            pos+=8;
        }
        else if( !strcmp( ltype, "Pos" ) ) {
            light_on[i] = 3;
            light_switches[POSITIONAL] = 1;
            lights[POSITIONAL].point.x = stof( argv[pos+1] );
            lights[POSITIONAL].point.y = stof( argv[pos+2] );
            lights[POSITIONAL].point.z = stof( argv[pos+3] );
            lights[POSITIONAL].attenuation1 = stof( argv[pos+4] );
            lights[POSITIONAL].attenuation2 = stof( argv[pos+5] );
            SET_COLOR( stof(argv[pos+7]), stof(argv[pos+8]), stof(argv[pos+9]),
                        lights[POSITIONAL].color.value );
            pos+=10;
        }
        else if( !strcmp( ltype, "Spt" ) ) {
            light_on[i] = 4;
            light_switches[SPOT] = 1;
            lights[SPOT].point.x = stof( argv[pos+1] );
            lights[SPOT].point.y = stof( argv[pos+2] );
            lights[SPOT].point.z = stof( argv[pos+3] );
            lights[SPOT].direction.x = stof( argv[pos+4] );
            lights[SPOT].direction.y = stof( argv[pos+5] );
            lights[SPOT].direction.z = stof( argv[pos+6] );
            NORMALIZE(lights[SPOT].direction);
            lights[SPOT].attenuation1 = stof( argv[pos+7] );
            lights[SPOT].attenuation2 = stof( argv[pos+8] );
            lights[SPOT].concentration = stof( argv[pos+9] );
            lights[SPOT].spread_angle = stof( argv[pos+10] );
            SET_COLOR( stof(argv[pos+12]),stof(argv[pos+13]),stof(argv[pos+14]),
                        lights[SPOT].color.value );
            pos+=15;
        }
    }

    num_off = stof( argv[pos+2] );
    pos += 3;
    for( i=0; i<num_off; i++ ) {
        sscanf( argv[pos], "%s", &ltype );
        if( !strcmp( ltype, "Amb" ) ) {
            light_off[i] = 1;
            SET_COLOR( stof(argv[pos+2]), stof(argv[pos+3]), stof(argv[pos+4]),
                        lights[AMBIENT].color.value );
            pos+=5;
        }
        else if( !strcmp( ltype, "Dir" ) ) {
            light_off[i] = 2;
            lights[DIRECTIONAL].direction.x = stof( argv[pos+1] );
            lights[DIRECTIONAL].direction.y = stof( argv[pos+2] );
            lights[DIRECTIONAL].direction.z = stof( argv[pos+3] );
            NORMALIZE(lights[DIRECTIONAL].direction);
            SET_COLOR( stof(argv[pos+5]), stof(argv[pos+6]), stof(argv[pos+7]),
                        lights[DIRECTIONAL].color.value );
            pos+=8;
        }
        else if( !strcmp( ltype, "Pos" ) ) {
            light_off[i] = 3;
            lights[POSITIONAL].point.x = stof( argv[pos+1] );
            lights[POSITIONAL].point.y = stof( argv[pos+2] );
            lights[POSITIONAL].point.z = stof( argv[pos+3] );
            lights[POSITIONAL].attenuation1 = stof( argv[pos+4] );
            lights[POSITIONAL].attenuation2 = stof( argv[pos+5] );
            SET_COLOR( stof(argv[pos+7]), stof(argv[pos+8]), stof(argv[pos+9]),
                        lights[POSITIONAL].color.value );
            pos+=10;
        }
        else if( !strcmp( ltype, "Spt" ) ) {
            light_off[i] = 4;
            lights[SPOT].point.x = stof( argv[pos+1] );
            lights[SPOT].point.y = stof( argv[pos+2] );
            lights[SPOT].point.z = stof( argv[pos+3] );
            lights[SPOT].direction.x = stof( argv[pos+4] );
            lights[SPOT].direction.y = stof( argv[pos+5] );
            lights[SPOT].direction.z = stof( argv[pos+6] );
            NORMALIZE(lights[SPOT].direction);
            lights[SPOT].attenuation1 = stof( argv[pos+7] );
            lights[SPOT].attenuation2 = stof( argv[pos+8] );
            lights[SPOT].concentration = stof( argv[pos+9] );
            lights[SPOT].spread_angle = stof( argv[pos+10] );
            SET_COLOR( stof(argv[pos+12]),stof(argv[pos+13]),stof(argv[pos+14]),
                        lights[SPOT].color.value );
            pos+=15;
        }
    }
         if( !strcmp( argv[pos+1], "Non" ) )
        cur_shading = PEXSurfaceInterpNone;
    else if( !strcmp( argv[pos+1], "Clr" ) )
        cur_shading = PEXSurfaceInterpColor;
    else if( !strcmp( argv[pos+1], "Dot" ) )
        cur_shading = PEXSurfaceInterpDotProduct;
    else if( !strcmp( argv[pos+1], "Nor" ) )
        cur_shading = PEXSurfaceInterpNormal;
         if( !strcmp( argv[pos+3], "Non" ) )
        cur_refl = PEXReflectionNone;
    else if( !strcmp( argv[pos+3], "Amb" ) )
        cur_refl = PEXReflectionAmbient;
    else if( !strcmp( argv[pos+3], "Dif" ) )
        cur_refl = PEXReflectionDiffuse;
    else if( !strcmp( argv[pos+3], "Spc" ) )
        cur_refl = PEXReflectionSpecular;
    }

    /* Initialize the toolkit. */
    toplevel = XtVaAppInitialize( &app, "XLightEditor",
        (XrmOptionDescList)NULL, 0,
        (int *)&argc, (String *)argv,
        (String *)NULL, NULL );

    /* See what display is being used. */
    dpy = XtDisplay( toplevel );
    XSetErrorHandler( PrintDefaultError );

    /* See if a PEX extension exists and initialize PEXlib. */
    if ( PEXInitialize( dpy, &pexinfo, 80, err ) != 0 ) {
        fprintf( stderr, "No PEX extension.\n" );
        fprintf( stderr, "%s\n", err );
        exit(1);
    }

    /* Determine the best visual to use. */
    ora_find_best_visual( dpy, &vis_info );

    if( vis_info.depth != 8 )
        ora_create_colormap( dpy, &vis_info, &cmap_info, &capx_info );
    else {
        /* Get a standard colormap for the visual. */
        if ( !ora_get_standard_colormap( dpy, &vis_info, &cmap_info ) ) {
            fprintf ( stderr, "Cannot find a standard colormap\n" );
            exit(1);
        }
        ora_set_stdcmap_approx( &vis_info, &cmap_info, &capx_info );
    }

    if( !(strcmp( pexinfo->vendor_name, "International Business Machines" ) ) )
        XtVaSetValues( toplevel,
            XtNvisual, vis_info.visual,
            XtNdepth, vis_info.depth,
            XtNcolormap, cmap_info.colormap,
            NULL );

    create_gui( vis_info.depth, pexinfo->vendor_name );
    XtRealizeWidget( toplevel );

    if( !(strcmp( pexinfo->vendor_name, "International Business Machines" ) ) )
        window = XtWindow( canvas );
    else {
        background.pixel = 0;
        background.flags = DoRed | DoGreen | DoBlue;
        window = ora_create_window( dpy, XtWindow(canvas), &vis_info,&cmap_info,
            "210x210", &background );
    }

    if( PEXUtDBConfig( dpy, window, True, &background, False, &drawable ) ) {
        fprintf( stderr, "No Double Buffering Support\n" );
        drawable = window;
    }

    /* Create the PEX renderer. */
    if ( IMMED_MODE_SPT( pexinfo ) ) {
        renderer = ora_setup_renderer( dpy, drawable, &capx_info, &rattrs );
    } else {
        fprintf( stderr, "Immediate mode not supported.\n" );
        exit(1);
    }

    pcMask[0] = pcMask[1] = pcMask[2] = 0;
    pcattrs.view_index           = 1;
    PEXSetPCAttributeMask( pcMask, PEXPCViewIndex );
    pcattrs.depth_cue_index      = 1;
    PEXSetPCAttributeMask( pcMask, PEXPCDepthCueIndex );
    pcattrs.distinguish          = True;
    PEXSetPCAttributeMask( pcMask, PEXPCDistinguishFlag );
    pcattrs.interior_style       = PEXInteriorStyleSolid;
    PEXSetPCAttributeMask( pcMask, PEXPCInteriorStyle );
    pcattrs.bf_interior_style    = PEXInteriorStyleSolid;
    PEXSetPCAttributeMask( pcMask, PEXPCBFInteriorStyle );
    pcattrs.surface_color.type   = PEXColorTypeRGB;
    SET_COLOR( 1, 0.2, 0.5, pcattrs.surface_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCSurfaceColor );
    pcattrs.bf_surface_color.type= PEXColorTypeRGB;
    SET_COLOR( 0.5, 0.5, 0.5, pcattrs.bf_surface_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCBFSurfaceColor );
    pcattrs.reflection_attr.ambient = 0.4;
    pcattrs.reflection_attr.diffuse = 0.4;
    pcattrs.reflection_attr.specular = 0.5;
    pcattrs.reflection_attr.specular_conc = 3.0;
    pcattrs.reflection_attr.transmission = 0;
    pcattrs.reflection_attr.specular_color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, pcattrs.reflection_attr.specular_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCReflectionAttr );
    pcattrs.bf_reflection_attr   = pcattrs.reflection_attr;
    PEXSetPCAttributeMask( pcMask, PEXPCBFReflectionAttr );
    rattrs.pipeline_context =
    PEXCreatePipelineContext( dpy, pcMask, &pcattrs );
    PEXChangeRenderer( dpy, renderer, PEXRAPipelineContext, &rattrs );

    colorTypeIndexed = ora_enum_type_supported( dpy, drawable,
        PEXETColorType, PEXColorTypeIndexed );
    colorTypeRGB = ora_enum_type_supported( dpy, drawable,
        PEXETColorType, PEXColorTypeRGB );
    colorTypeHSV = ora_enum_type_supported( dpy, drawable,
        PEXETColorType, PEXColorTypeHSV );
    colorTypeHLS = ora_enum_type_supported( dpy, drawable,
        PEXETColorType, PEXColorTypeHLS );
    colorTypeCIE = ora_enum_type_supported( dpy, drawable,
        PEXETColorType, PEXColorTypeCIE );

    build_sphere();

    top_struct = PEXCreateStructure( dpy );
#if( 0 )    /* moved into pc */
    PEXSetViewIndex( dpy, top_struct, PEXOCStore, 1 );
    PEXSetDepthCueIndex( dpy, top_struct, PEXOCStore, 1 );

    PEXSetFacetDistinguishFlag( dpy, top_struct, PEXOCStore, True );
    PEXSetInteriorStyle( dpy, top_struct, PEXOCStore, PEXInteriorStyleSolid );
    PEXSetBFInteriorStyle( dpy, top_struct, PEXOCStore, PEXInteriorStyleSolid );

    SET_COLOR( 1, 0.2, 0.5, color );
    PEXSetSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeRGB, &color );
    SET_COLOR( 0.5, 0.5, 0.5, color );
    PEXSetBFSurfaceColor( dpy, top_struct, PEXOCStore, PEXColorTypeRGB, &color);
#endif		/*  */
    PEXLabel( dpy, top_struct, PEXOCStore, LIGHT_STATE_LABEL );
    PEXSetLightSourceState( dpy, top_struct, PEXOCStore,
        num_on, light_on, num_off, light_off );

    PEXLabel( dpy, top_struct, PEXOCStore, REFL_EQN_LABEL );
    PEXSetReflectionModel( dpy, top_struct, PEXOCStore, cur_refl );
    PEXSetBFReflectionModel( dpy, top_struct, PEXOCStore, cur_refl );
#if( 0 )    /* moved into pc */
    refl_props.ambient = 0.4;
    refl_props.diffuse = 0.4;
    refl_props.specular = 0.5;
    refl_props.specular_conc = 3.0;
    refl_props.transmission = 0;
    refl_props.specular_color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, refl_props.specular_color.value );
    PEXSetReflectionAttributes( dpy, top_struct, PEXOCStore, &refl_props );
    PEXSetBFReflectionAttributes( dpy, top_struct, PEXOCStore,  &refl_props );
#endif		/*  */
    PEXLabel( dpy, top_struct, PEXOCStore, SHADING_METHOD_LABEL );
    PEXSetSurfaceInterpMethod( dpy, top_struct, PEXOCStore, cur_shading );
    PEXSetBFSurfaceInterpMethod( dpy, top_struct, PEXOCStore, cur_shading );

    PEXLabel( dpy, top_struct, PEXOCStore, OBJECT_LABEL );
    PEXExecuteStructure( dpy, top_struct, PEXOCStore, sphere );

    /* Initialize view 1. */
    initialize_view();

    /* Initialize light table and depth cueing table. */
    PEXSetTableEntries( dpy, rattrs.light_table, 1, 1,
        PEXLUTLight, &lights[AMBIENT] );
    PEXSetTableEntries( dpy, rattrs.light_table, 2, 1,
        PEXLUTLight, &lights[DIRECTIONAL] );
    PEXSetTableEntries( dpy, rattrs.light_table, 3, 1,
        PEXLUTLight, &lights[POSITIONAL] );
    PEXSetTableEntries( dpy, rattrs.light_table, 4, 1,
        PEXLUTLight, &lights[SPOT] );

    PEXLabel( dpy, top_struct, PEXOCStore, START_LIGHT_LABEL );

    PEXLabel( dpy, top_struct, PEXOCStore, DIR_LIGHT_LABEL );
    if( islighton(2) ) {
        pln[0].x = (lights[DIRECTIONAL].direction.x<0.) ? -lights[DIRECTIONAL].direction.x+1. : -lights[DIRECTIONAL].direction.x-1.;
        pln[0].y = (lights[DIRECTIONAL].direction.y<0.) ? -lights[DIRECTIONAL].direction.y+1. : -lights[DIRECTIONAL].direction.y-1.;
        pln[0].z = (lights[DIRECTIONAL].direction.z<0.) ? -lights[DIRECTIONAL].direction.z+1. : -lights[DIRECTIONAL].direction.z-1.;
        pln[0].x = (lights[DIRECTIONAL].direction.x==0.) ? 0. : pln[0].x;
        pln[0].y = (lights[DIRECTIONAL].direction.y==0.) ? 0. : pln[0].y;
        pln[0].z = (lights[DIRECTIONAL].direction.z==0.) ? 0. : pln[0].z;
        pln[1].x = -lights[DIRECTIONAL].direction.x;
        pln[1].y = -lights[DIRECTIONAL].direction.y;
        pln[1].z = -lights[DIRECTIONAL].direction.z;
        PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
    }
    else
        PEXPolyline( dpy, top_struct, PEXOCStore, 0, (PEXCoord *)0 );

    PEXLabel( dpy, top_struct, PEXOCStore, POS_LIGHT_LABEL );
    if( islighton(3) ) {
        PEXMarkers( dpy, top_struct, PEXOCStore, 1, &lights[POSITIONAL].point );
    }
    else
        PEXMarkers( dpy, top_struct, PEXOCStore, 0, (PEXCoord *)0 );

    PEXLabel( dpy, top_struct, PEXOCStore, SPT_LIGHT_LABEL );
    if( islighton(4) ) {
        pln[0].x = lights[SPOT].point.x;
        pln[0].y = lights[SPOT].point.y;
        pln[0].z = lights[SPOT].point.z;
        pln[1].x = lights[SPOT].point.x + lights[SPOT].direction.x;
        pln[1].y = lights[SPOT].point.y + lights[SPOT].direction.y;
        pln[1].z = lights[SPOT].point.z + lights[SPOT].direction.z;
        PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
    }
    else
        PEXPolyline( dpy, top_struct, PEXOCStore, 0, (PEXCoord *)0 );

    PEXLabel( dpy, top_struct, PEXOCStore, END_LIGHT_LABEL );

    initialize_depth_cue_table();

    redraw();
    for (;;) {
        XEvent event;
        XtAppNextEvent(app, &event);
        if( (event.type == Expose) &&
            (event.xexpose.window == window) &&
            (event.xexpose.count == 0) )
            redraw();
        else
            XtDispatchEvent(&event);
    }
}


float stof( buf )
char *buf;
{
    float f;
    sscanf( buf, "%f", &f );
    return f;
}


void
redraw()
{
    /* Redraw the window, after first erasing (clearing) it. */
    PEXRenderNetwork( dpy, drawable, renderer, top_struct );
    drawable = PEXUtDBSwap( dpy, window, False );
}


static void
initialize_light_table()
{
    lights[AMBIENT].type = PEXLightAmbient;
    lights[AMBIENT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[AMBIENT].color.value );

    lights[DIRECTIONAL].type = PEXLightWCVector;
    lights[DIRECTIONAL].direction.x = -1;
    lights[DIRECTIONAL].direction.y =  0;
    lights[DIRECTIONAL].direction.z =  0;
    NORMALIZE(lights[DIRECTIONAL].direction);
    lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[DIRECTIONAL].color.value );

    lights[POSITIONAL].type = PEXLightWCPoint;
    lights[POSITIONAL].point.x = 1.0;
    lights[POSITIONAL].point.y = 2.0;
    lights[POSITIONAL].point.z = 0.7;
    lights[POSITIONAL].attenuation1 = 0;
    lights[POSITIONAL].attenuation2 = 1;
    lights[POSITIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[POSITIONAL].color.value );

    lights[SPOT].type = PEXLightWCSpot;
    lights[SPOT].point.x = 0.5;
    lights[SPOT].point.y = 0.5;
    lights[SPOT].point.z = 2.0;
    lights[SPOT].direction.x = 0;
    lights[SPOT].direction.y = 0;
    lights[SPOT].direction.z = 1;
    NORMALIZE(lights[SPOT].direction);
    lights[SPOT].attenuation1 = 0.5;
    lights[SPOT].attenuation2 = 0.1;
    lights[SPOT].concentration = 0;
    lights[SPOT].spread_angle = M_PI/2;
    lights[SPOT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[SPOT].color.value );
}

void
initialize_depth_cue_table()
{
    depth_cue_rep.mode = PEXOff;
    depth_cue_rep.front_plane = 1;
    depth_cue_rep.back_plane = 0;
    depth_cue_rep.front_scaling = 1;
    depth_cue_rep.back_scaling = 0.5;
    depth_cue_rep.color.type = PEXColorTypeRGB;
    SET_COLOR( 0, 0, 0, depth_cue_rep.color.value );
    PEXSetTableEntries( dpy, rattrs.depth_cue_table, 1, 1,
        PEXLUTDepthCue, &depth_cue_rep );
}

void
initialize_view()
{
    int            error;

    /* Compute the view orientation matrix. */
    error = PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    /* Set the view mapping parameters. */
    perspective = True;
    prp.x = 0; 
    prp.y = 0; 
    prp.z = 10.0;
    view_window[0].x = -1.5; 
    view_window[1].x = 1.5;
    view_window[0].y = -1.5; 
    view_window[1].y = 1.5;
    back_plane = -2.0; 
    front_plane = 2.0; 
    view_plane = 0.0;
    viewport.min.x = 0.0;    viewport.max.x = 1.0;
    viewport.min.y = 0.0;    viewport.max.y = 1.0;
    viewport.min.z = 0.0;    viewport.max.z = 1.0;
    error = PEXViewMappingMatrix( view_window, &viewport, perspective,
        &prp, view_plane, back_plane,
        front_plane, view.mapping );

    /* Set the view representation. */
    view.clip_flags = PEXClippingAll;
    view.clip_limits = viewport;
    PEXSetTableEntries( dpy, rattrs.view_table, 1, 1, PEXLUTView, &view );
}

#define NUM_LONG_FACETS   32
#define NUM_LONG_POINTS   (NUM_LONG_FACETS + 1)
#define NUM_LAT_FACETS    16
#define NUM_LAT_POINTS    (NUM_LAT_FACETS + 1)
#define RADIUS            0.95

void
build_sphere()
{
    PEXVertexNormal    vertex_data[NUM_LONG_POINTS * NUM_LAT_POINTS];
    PEXArrayOfVertex    vdata;
    PEXArrayOfFacetData    facet_data;
    PEXVector        facet_normals[NUM_LONG_FACETS * NUM_LAT_FACETS];
    unsigned int    facet_data_flag, vertex_data_flag;
    double        theta = 0, phi = 0;
    register int  i, j, index;

    facet_data_flag = PEXGANormal;
    vertex_data_flag = PEXGANormal;

    for ( i = 0; i < NUM_LAT_POINTS; i++ ) {
        for ( j = 0; j < NUM_LONG_FACETS; j++ ) {
            index = i * NUM_LONG_POINTS + j;
            vertex_data[index].point.x = RADIUS * (1 + sin(phi) * cos(theta));
            vertex_data[index].point.y = RADIUS * (1 + cos(phi));
            vertex_data[index].point.z = RADIUS * (1 + sin(phi) * sin(theta));
            vertex_data[index].normal.x = sin(phi) * cos(theta);
            vertex_data[index].normal.y = cos(phi);
            vertex_data[index].normal.z = sin(phi) * sin(theta);
            NORMALIZE(vertex_data[index].normal);
            theta += (2 * M_PI) / NUM_LONG_FACETS;
        }
        phi += M_PI / NUM_LAT_FACETS;

        /* Close the circle. */
        index = i * NUM_LONG_POINTS + NUM_LONG_FACETS;
        vertex_data[index].point = vertex_data[index - NUM_LONG_FACETS].point;
        vertex_data[index].normal =
            vertex_data[index - NUM_LONG_FACETS].normal;
    }

    theta = M_PI / NUM_LONG_FACETS;
    phi = 0.5 * M_PI / NUM_LAT_FACETS;
    index = 0;
    for ( i = 0; i < NUM_LAT_FACETS; i++ ) {
        for ( j = 0; j < NUM_LONG_FACETS; j++ ) {
            facet_normals[index].x = sin(phi) * cos(theta);
            facet_normals[index].y = cos(phi);
            facet_normals[index].z = sin(phi) * sin(theta);
            ++index;
            theta += (2 * M_PI) / NUM_LONG_FACETS;
        }
        phi += M_PI / NUM_LAT_FACETS;
    }

    sphere = PEXCreateStructure( dpy );
    PEXSetFacetCullingMode( dpy, sphere, PEXOCStore, PEXBackFaces );
    vdata.normal = vertex_data;
    facet_data.normal = facet_normals;
    PEXQuadrilateralMesh( dpy, sphere, PEXOCStore, PEXShapeUnknown,
        facet_data_flag, vertex_data_flag,
        PEXColorTypeRGB, facet_data,
        NUM_LONG_POINTS, NUM_LAT_POINTS, vdata );
}
